# @Author: Bi Ying
# @Date:   2024-07-26 23:48:04
from typing import List, Dict, Optional

from pydantic import BaseModel, Field
from openai.types.chat.chat_completion_message import ChatCompletionMessageToolCall

from . import defaults as defs


class EndpointSetting(BaseModel):
    id: str = Field(..., description="The id of the endpoint.")
    region: Optional[str] = Field(None, description="The region for the endpoint.")
    api_base: str = Field(None, description="The base URL for the API.")
    api_key: Optional[str] = Field(None, description="The API key for authentication.")
    credentials: Optional[dict] = Field(None, description="Additional credentials if needed.")
    is_azure: bool = Field(False, description="Indicates if the endpoint is for Azure.")
    is_vertex: bool = Field(False, description="Indicates if the endpoint is for Vertex.")
    rpm: int = Field(description="Requests per minute.", default=defs.ENDPOINT_RPM)
    tpm: int = Field(description="Tokens per minute.", default=defs.ENDPOINT_TPM)
    concurrent_requests: int = Field(
        description="Whether to use concurrent requests for the LLM service.",
        default=defs.ENDPOINT_CONCURRENT_REQUESTS,
    )


class ModelSetting(BaseModel):
    id: Optional[str] = Field(None, description="The id of the model.")
    endpoints: List[str] = Field(default_factory=list, description="Available endpoints for the model.")
    function_call_available: bool = Field(False, description="Indicates if function call is available.")
    response_format_available: bool = Field(False, description="Indicates if response format is available.")
    context_length: int = Field(32768, description="The context length for the model.")
    max_output_tokens: Optional[int] = Field(None, description="Maximum number of output tokens allowed.")


class BackendSettings(BaseModel):
    models: Dict[str, ModelSetting] = Field(default_factory=dict)

    def update_models(self, default_models: Dict[str, Dict], input_models: Dict[str, Dict]):
        updated_models = {}
        for model_name, model_data in default_models.items():
            updated_model = ModelSetting(**model_data)
            if model_name in input_models:
                updated_model = updated_model.model_copy(update=input_models[model_name])
            updated_models[model_name] = updated_model

        # Add any new models from input that weren't in defaults
        for model_name, model_data in input_models.items():
            if model_name not in updated_models:
                updated_models[model_name] = ModelSetting(**model_data)

        self.models = updated_models


class Usage(BaseModel):
    completion_tokens: int
    prompt_tokens: int
    total_tokens: int


class ModelOutput(BaseModel):
    content: Optional[str] = None

    tool_calls: Optional[List[ChatCompletionMessageToolCall]] = None
    """The tool calls generated by the model, such as function calls."""

    function_call_arguments: Optional[dict] = None

    usage: Optional[Usage] = None
